#pragma implementation "TextEdit.h"
#include <ctype.h>
#include <keyboard.h>
#include <TextEdit.h>
#include "virtpos.h"

/* The ImageWin class from edximg could be used (modulo load/store),
 * and can display attributes, but it imposes a maximum line length.
 * Very long lines can be an efficiency problem with this model
 * because tabs are expanded for display.
 */

TextEditor::TextEditor(StringList *model):
  TextBrowser(view),
  view(model ? *model : *new StringList),
  txt(view.txt)
{
  txt[0] = '~';		// default string
  wrapmargin = 0;
}

TextEditor::~TextEditor() {
  delete &txt;
}

void TextEditor::key(int ky) {
  switch (ky) {
  case KEY_TAB:
    if (insert_mode) {
      insch('\t');
      return;
    }
    break;
  case KEY_ENTER:
    if (insert_mode) {
      int r = row + topline;
      string &line = txt[r];
      insertln(r + 1);
      int idx = virtpos(line.c_str()).mappos(leftcol + col);
      txt[++r] = line.substr(idx);
      // make enter a hard return by removing trailing space
      while (idx && isspace(line[idx - 1]))
	--idx;
      line.erase(idx);
      move(r,0);
      draw(Rect(row - 1,0,row,loc.c2));
      reform();
      return;
    }
    break;
  case KEY_BACKSPACE:
    if (leftcol + col <= 0) {
      if (row + topline > 0) {
	int r = row + topline;
	move(r,-1);
	txt[r-1] += txt[r];
	draw(Rect(row,col,row,loc.c2));
	deleteln(r);
	reform();
      }
      return;
    }
    break;
  }
  EditWin::key(ky);
}

void TextEditor::addch(char c) {
  int r = row + topline;
  if (r >= txt.size())
    txt.add(string(&c,1));
  else {
    string &line = txt[r];
    virtpos p(line.c_str());
    if (p.mappos(leftcol + col) < line.length())
      line[p.idx] = c;
    else {
      insch(c);
      return;
    }
  }
  draw(Rect(row,col,row,loc.c2));
  move(r,leftcol + col + 1);
  modified = 1;
}

void TextEditor::insch(char c) {
  int r = row + topline;
  if (r >= txt.size())
    txt.add(string(&c,1));
  else {
    string &line = txt[r];
    virtpos p(line.c_str());
    if (p.mappos(leftcol + col) >= line.length()) {
      int pos = p.pos;
      while (pos < leftcol + col) {
	line += ' ';
	virtpos p1(line.c_str());
	p1.mappos(leftcol + col);
	pos = p1.pos;
      }
      line += c;
    }
    else
      line.insert(p.idx,1,c);
  }
  move(r,leftcol + col + 1);
  reform(TRUE);
  /* draw(Rect(row,col,row,loc.c2)); */
  modified = 1;
}

void TextEditor::delch() {
  string &line = txt[row + topline];
  int idx = virtpos(line.c_str()).mappos(leftcol + col);
  if (idx < line.length()) {
    line.erase(idx,1);
    /* draw(Rect(row,col,row,loc.c2)); */
    reform();
    modified = 1;
  }
}

void TextEditor::insertln(int r) {
  txt.insertln(r);
  slide(Rect(r - topline,0,loc.height(),loc.c2),DIR_DOWN,1);
  modified = 1;
}

void TextEditor::deleteln(int r) {
  txt.deleteln(r);
  slide(Rect(r - topline + 1,0,loc.height(),loc.c2),DIR_UP,1);
  modified = 1;
}

void TextEditor::clrtoeol() {
  string &line = txt[row + topline];
  line.erase(virtpos(line.c_str()).mappos(col + leftcol));
  draw(Rect(row,col,row,loc.c2));
  modified = 1;
}

static int wordlen(const char *s) {
  int i;
  for (i = 0; s[i] && !isspace(s[i]); ++i)
    ;
  return i;
}

static int wraplen(const string &str,int margin) {
  int i = str.length();
  const char *s = str.c_str();
  if (i > margin) i = margin;
  while (i > 0 && isspace(s[i-1]))
    --i;
  while (i > 0 && !isspace(s[i-1]))
    --i;
  return i;
}

void TextEditor::reform(bool hardret) {
  int r1 = row + topline;
  int r;
  for (r = r1; r < txt.size(); hardret = FALSE) {
    string &line = txt[r++];
    if (wrapmargin <= wordlen(line.c_str())) break;
    virtpos p(line.c_str());
    int len = line.length();
    if (r >= txt.size() || !len || !isspace(line[len - 1]))
      hardret = TRUE;
    int mlen = p.mappos(wrapmargin);
    if (len > mlen) {
      int w = wraplen(line,mlen);
      if (hardret)
	insertln(r);
      string &cont = txt[r];
      cont = line.substr(w) + cont;
      int c = p.mappos(leftcol + col);
      line.erase(w);
      if (c >= w && r - 1 == row + topline) {
	/* move(r,0); */
	move(r,c - w);
      }
      continue;
    }
    string &nxtl = txt[r];
    if (hardret || p.mapidx(len) + wordlen(nxtl.c_str()) >= wrapmargin) break;
    if (nxtl.length() == 0) break;
    line += nxtl;
    deleteln(r--);
  }
  draw(Rect(r1 - topline,0,r - topline - 1,loc.c2));
}

void TextEditor::cmd(int c) {
  switch (c) {
  case EDIT_REFORM:
    reform();
    setfocus(this);
    return;
  }
  EditWin::cmd(c);
}
